/* ENGLISH.CMD: Install PLUMA in English. */

'@Echo Off'

/* Set the installed object's title. */

Title = 'Pluma 1.12'


/* Load REXXUTIL */

Call RxFuncAdd 'SysLoadFuncs', 'REXXUTIL', 'SysLoadFuncs'
Call SysLoadFuncs


/* Initialize */
 
Signal On Failure Name FAILURE
Signal On Halt Name HALT
Signal On Syntax Name SYNTAX

Call SysCls
Say 'Installing PLUMA...'
Say ''


/* Verify the existence of the various component files. */

Language = 'ENGLISH'

Result = SysFileTree( 'PLUMA.EXE', 'Files', 'F' )
If Files.0 = 0 Then
   Do
   Say 'ERROR: PLUMA.EXE not found!'
   Signal DONE
   End

Result = SysFileTree( Language".DLL", 'Files', 'F' )
If Files.0 = 0 Then
   Do
   Say 'ERROR: 'Language'.DLL not found!'
   Signal DONE
   End

Result = SysFileTree( Language".HLP", 'Files', 'F' )
If Files.0 = 0 Then
   Do
   Say 'ERROR: 'Language'.HLP not found!'
   Signal DONE
   End

/* Ask for the target directory name. */
 
Say 'Please enter the full name of the directory to which'
Say '  you want PLUMA installed (default C:\OS2\APPS): '
Pull Directory
If Directory = "" Then Directory = 'C:\OS2\APPS'


/* Create the target directory if necessary. */

Result = SysFileTree( Directory, 'Dirs', 'D' )
If Dirs.0 = 0 Then
   Do
   Result = SysMkDir( Directory )
   if Result == 0 Then
      Do
      End
   Else
      Do
      Say 'ERROR: Unable to create target directory.'
      Signal DONE
      End
   End
Say ''


/* Perform the installation. */

Say 'Copying PLUMA to ' Directory '...'
Copy PLUMA.EXE Directory                   '1>NUL'
Copy PLUMA.ICO Directory                   '1>NUL'
Copy Language".DLL" Directory"\PLUMA.DLL"  '1>NUL'
Copy Language".HLP" Directory"\PLUMA.HLP"  '1>NUL'
Copy Language".DOC" Directory"\PLUMA.DOC"  '1>NUL'

Result = SysFileTree( Language".INF", 'Files', 'F' )
If Files.0 > 0 Then
   Do
   Copy Language".INF" Directory"\PLUMA.INF"  '1>NUL'
   End

Say 'Creating program folder...'
Type = 'WPFolder'
Folder = '<WP_DESKTOP>'
Parms = "OPEN=ICON;OBJECTID=<PLUMAF>;ICONFILE="Directory"\PLUMA.ICO;"
Result = SysCreateObject( Type, 'Pluma', Folder, Parms, 'ReplaceIfExists' )
If Result != 1 Then
   Say 'ERROR: Folder not created.'

Say 'Creating program object...'
Type = 'WPProgram'
Folder = '<PLUMAF>'
Parms = "MINWIN=DESKTOP;PROGTYPE=PM;EXENAME="Directory"\PLUMA.EXE;STARTUPDIR="Directory";OBJECTID=<PLUMA>;NOPRINT=YES;ASSOCTYPE=Pluma;"
Result = SysCreateObject( Type, Title, Folder, Parms, 'ReplaceIfExists' )
If Result != 1 Then
   Say 'ERROR: Object not created.'

Say 'Creating documentation objects...'
Type = 'WPShadow'
Parms = "SHADOWID="Directory"\PLUMA.DOC;OBJECTID=<PLUMAD>;"
Result = SysCreateObject( Type, Title, Folder, Parms, 'ReplaceIfExists' )
If Result != 1 Then
   Say 'ERROR: Shadow not created.'

Result = SysFileTree( Directory"\PLUMA.INF", 'Files', 'F' )
If Files.0 > 0 Then Do
  Parms = "SHADOWID="Directory"\PLUMA.INF;OBJECTID=<PLUMAI>;"
  Result = SysCreateObject( Type, Title, Folder, Parms, 'ReplaceIfExists' )
  If Result != 1 Then Do
    Say 'ERROR: Shadow not created.'
    End
  End
 
Signal DONE

FAILURE:
Say 'REXX failure.'
Signal DONE

HALT:
Say 'REXX halt.'
Signal DONE

SYNTAX:
Say 'REXX syntax error.'
Signal DONE

DONE:
Exit
